from .desc import Description, Value, SideEffect
from .guard import make_guard, Guard
from .parse import parse_func
from .mark import (
    mark_input, mark_output, mark_side_effect,
    Val, Outputs,
)

__all__ = [
    "Description", "Value", "SideEffect",
    "make_guard", "Guard",
    "parse_func",
    "mark_input", "mark_output", "mark_side_effect",
    "Val", "Outputs"
]

__version__ = '0.1.1'



import typing as T
from pathlib import Path

from ..desc import Value


class ValueType(object):
    """The base custom value type, for
    custom type and range checking.

    Sub-class should override the `check_type` or
    `check_range` method.
    """
    @staticmethod
    def check_type(val: T.Any, type_: T.Type) -> bool:
        return True

    @staticmethod
    def check_range(val: T.Any, range_: T.Any) -> bool:
        return True


class OneOf(ValueType):
    """Input value should be equal to at least one elements in `range`"""
    @staticmethod
    def check_range(val, range_):
        return val in range_


class SubSet(ValueType):
    """Input value should be a subset of `range`"""
    @staticmethod
    def check_range(val, range_):
        return all([v in range_ for v in val])


class InputPath(ValueType):
    """Input value should be a file path,
    and the file should be exist."""
    @staticmethod
    def check_type(val, type_):
        return isinstance(val, str) or isinstance(val, Path)

    @staticmethod
    def check_range(val, range_):
        path = Path(val) if isinstance(val, str) else val
        return path.exists()


class OutputPath(ValueType):
    """Input value should be a file path. """
    @staticmethod
    def check_type(val, type_):
        return isinstance(val, str) or isinstance(val, Path)


Value.register_range_check(OneOf, OneOf.check_range)
Value.register_range_check(SubSet, SubSet.check_range)
Value.register_type_check(InputPath, InputPath.check_type)
Value.register_range_check(InputPath, InputPath.check_range)
Value.register_type_check(OutputPath, OutputPath.check_type)



import types
import typing as T
import inspect

from .desc import Description, Value
from .mark import FUNC_MARK_STORE_KEY, FuncMarks


GenericAlias = T._GenericAlias  # type: ignore


def _update_val_by_marks(
        mark_idx: int, name: str,
        marks: T.Dict, val: Value
        ):
    if mark_idx in marks:
        store = marks[mark_idx].copy()
    elif name in marks:
        store = marks[name].copy()
    else:
        store = None

    if store is not None:
        val.kwargs.update(store.pop("attrs"))
        val.__dict__.update(store)


def parse_func_inputs(
        sig: inspect.Signature,
        is_method: bool,
        marks: T.Dict[T.Union[int, str], dict],
        ) -> T.List[Value]:
    inputs = []
    for idx, (name, param) in enumerate(sig.parameters.items()):
        ann = param.annotation
        if isinstance(ann, Value):
            val = ann
        elif ann is inspect._empty:
            val = Value(None)
        else:
            val = Value(ann)

        val.name = name

        # set default value
        if param.default is not inspect._empty:
            val.default = param.default

        # update by marks
        mark_idx = idx + 1 if is_method else idx
        _update_val_by_marks(mark_idx, name, marks, val)

        inputs.append(val)
    return inputs


def parse_func_outputs(
        sig: inspect.Signature,
        marks: T.Dict[T.Union[int, str], dict]
        ) -> T.List[Value]:
    outputs = []
    ret = sig.return_annotation

    def to_val(o):
        return o if isinstance(o, Value) else Value(o)

    val: Value
    if ret is inspect._empty:
        val = Value(type(None))
        outputs.append(val)
    elif isinstance(ret, Value):
        outputs.append(ret)
    elif isinstance(ret, list):
        outputs.extend([to_val(o) for o in ret])
    elif isinstance(ret, GenericAlias) and \
            (ret._name == "Tuple"):
        outputs.extend([to_val(o) for o in ret.__args__])
    else:
        val = Value(ret)
        outputs.append(val)

    # update by marks
    for idx, val in enumerate(outputs):
        if val.name == "?":
            val.name = f"output_{idx}"

        _update_val_by_marks(idx, val.name, marks, val)
    return outputs


def parse_func(func: T.Callable) -> Description:
    desc = Description()
    sig = inspect.signature(func)
    is_method = isinstance(func, types.MethodType)
    func_marks: T.Optional[FuncMarks] = func.__dict__.get(FUNC_MARK_STORE_KEY)
    if func_marks is None:
        func_marks = FuncMarks()
    inputs = parse_func_inputs(sig, is_method, func_marks.input_marks)
    desc.inputs = inputs
    outputs = parse_func_outputs(sig, func_marks.output_marks)
    desc.outputs = outputs
    desc.side_effects = func_marks.side_effect_marks
    return desc



import typing as T
import functools
import types

from .desc import Value, Description
from .parse import parse_func


class CheckError(Exception):
    pass


class SideEffectError(Exception):
    pass


TF2 = T.TypeVar("TF2", bound=T.Callable)


class Guard(T.Generic[TF2]):
    def __init__(
            self,
            func: TF2,
            desc: T.Optional[Description] = None,
            check_inputs: bool = True,
            check_outputs: bool = True,
            check_side_effect: bool = False,
            check_type: bool = True,
            check_range: bool = True,
            ) -> None:
        self.func = func
        functools.update_wrapper(self, func)
        if desc is None:
            desc = parse_func(func)
        self.desc = desc
        self.is_check_type = check_type
        self.is_check_range = check_range
        self.is_check_side_effect = check_side_effect
        self.is_check_inputs = check_inputs
        self.is_check_outputs = check_outputs

    def __call__(self, *args, **kwargs):
        pass_in = self.desc.parse_pass_in(args, kwargs)
        errors = []
        if self.is_check_inputs:
            self.check_inputs(pass_in, errors)
        if self.is_check_side_effect:
            self.check_side_effects_before_run(
                pass_in, errors)
        res = self.func(*args, **kwargs)
        if self.is_check_outputs:
            self.check_outputs(res, errors)
        if self.is_check_side_effect:
            self.check_side_effects_after_run(
                pass_in, res, errors)
        return res

    def check_inputs(self, pass_in: dict, errors: list):
        for val in self.desc.inputs:
            self.check_value(val, pass_in[val.name], errors)
        if len(errors) > 0:
            raise CheckError(errors)

    def check_outputs(self, res: T.Union[T.Any, T.Tuple], errors: list):
        if isinstance(res, tuple):
            if len(res) != len(self.desc.outputs):
                raise CheckError(
                    f"Output num({len(res)}) not match the"
                    f" description outputs num({len(self.desc.outputs)})"
                )
            for idx, val in enumerate(self.desc.outputs):
                self.check_value(val, res[idx], errors)
        else:
            if len(self.desc.outputs) != 1:
                raise CheckError(
                    "Output num(1) not match the"
                    f" description outputs num({len(self.desc.outputs)})"
                )
            val = self.desc.outputs[0]
            self.check_value(val, res, errors)
        if len(errors) > 0:
            raise CheckError(errors)

    def check_value(
            self,
            arg: Value,
            val: T.Any,
            errors: T.List[Exception]):
        try:
            if self.is_check_type:
                arg.check_type(val)
            if self.is_check_range:
                arg.check_range(val)
        except Exception as e:
            errors.append(e)

    def get_input_dict(self, pass_in: dict) -> dict:
        in_dict: T.Dict[T.Union[int, str], T.Any] = {}
        for i, v in enumerate(self.desc.inputs):
            in_dict[i] = pass_in[v.name]
            in_dict[v.name] = pass_in[v.name]
        return in_dict

    def get_output_dict(self, res: T.Union[tuple, T.Any]) -> dict:
        rtn_dict: T.Dict[T.Union[int, str], T.Any]
        if self.is_check_outputs:
            if isinstance(res, tuple):
                rtn_dict = {}
                for i, v in enumerate(self.desc.outputs):
                    rtn_dict[i] = res[i]
                    rtn_dict[v.name] = res[i]
            else:
                rtn_dict = {
                    self.desc.outputs[0].name: res,
                    0: res,
                }
        else:
            rtn_dict = {}
        return rtn_dict

    def check_side_effects_before_run(
            self, pass_in: dict, errors: list):
        if len(self.desc.side_effects) == 0:
            return
        in_dict = self.get_input_dict(pass_in)
        for e in self.desc.side_effects:
            if not e.check_before_run(in_dict):
                err = SideEffectError(
                    "Error occured when check "
                    f"side effect(before run): {self.func}"
                )
                errors.append(err)
        if len(errors) > 0:
            raise CheckError(errors)

    def check_side_effects_after_run(
            self, pass_in: dict,
            res: T.Union[tuple, T.Any], errors: list):
        if len(self.desc.side_effects) == 0:
            return
        in_dict = self.get_input_dict(pass_in)
        rtn_dict = self.get_output_dict(res)
        for e in self.desc.side_effects:
            if not e.check_after_run(in_dict, rtn_dict):
                err = SideEffectError(
                    "Error occured when check "
                    f"side effect(after run): {self.func}"
                )
                errors.append(err)
        if len(errors) > 0:
            raise CheckError(errors)

    def __get__(self, obj, objtype):
        """Allow use on instance method."""
        if not hasattr(self, "_bounded"):  # bound only once
            target_func = self.func
            bound_mth = types.MethodType(target_func, obj)
            self.func = bound_mth
            self.desc.inputs.pop(0)
            self._bounded = True
        return self


def make_guard(
        func: T.Optional[TF2] = None,
        *,
        check_inputs: bool = True,
        check_outputs: bool = True,
        check_side_effect: bool = False,
        check_type: bool = True,
        check_range: bool = True,
        ) -> TF2:
    kwargs = {
        "check_inputs": check_inputs,
        "check_outputs": check_outputs,
        "check_side_effect": check_side_effect,
        "check_type": check_type,
        "check_range": check_range,
    }
    if func is None:
        return functools.partial(make_guard, **kwargs)  # type: ignore
    return Guard(func, **kwargs)  # type: ignore



import typing as T

from .desc import Value, _NotDef, NotDef, T1, SideEffect
from .utils.misc import CreateByGetItem


Val = Value


class Outputs(metaclass=CreateByGetItem):
    def __new__(
            cls,
            *outputs,
            ):
        return list(outputs)


FUNC_MARK_STORE_KEY = "_funcdesc_marks"


class FuncMarks():
    def __init__(self) -> None:
        self.input_marks: T.Dict[T.Union[str, int], T.Dict] = dict()
        self.output_marks: T.Dict[T.Union[str, int], T.Dict] = dict()
        self.side_effect_marks: T.List[SideEffect] = []


TF1 = T.TypeVar("TF1")


def _mark_val_factory(store_key: T.Literal["input", "output"]):
    def add_when_not_default(dict_, key, val, default):
        if val != default:
            dict_[key] = val

    def mark_val(
            pos_or_name: T.Union[int, str],
            *,
            type: T.Optional[T.Type[T1]] = None,
            range: T.Optional[T.Any] = None,
            default: T.Union[_NotDef, T1] = NotDef,
            name: T.Optional[str] = None,
            **attrs) -> T.Callable[[TF1], TF1]:

        store = {}
        store['attrs'] = attrs
        add_when_not_default(store, "type", type, None)
        add_when_not_default(store, "range", range, None)
        add_when_not_default(store, "default", default, NotDef)
        add_when_not_default(store, "name", name, None)

        def wrap(func: TF1) -> TF1:
            func.__dict__.setdefault(FUNC_MARK_STORE_KEY, FuncMarks())
            marks: FuncMarks = func.__dict__[FUNC_MARK_STORE_KEY]
            if store_key == "input":
                marks.input_marks[pos_or_name] = store
            else:
                marks.output_marks[pos_or_name] = store
            return func
        return wrap

    return mark_val


mark_input = _mark_val_factory("input")
mark_output = _mark_val_factory("output")


def mark_side_effect(side_effect: SideEffect):

    def wrap(func: T.Callable) -> T.Callable:
        func.__dict__.setdefault(FUNC_MARK_STORE_KEY, FuncMarks())
        marks: FuncMarks = func.__dict__[FUNC_MARK_STORE_KEY]
        marks.side_effect_marks.append(side_effect)
        return func

    return wrap



import re
from pathlib import Path

from ..desc import SideEffect


class WriteFile(SideEffect):
    def __init__(self, path_template: str):
        self.path_template = path_template

    @property
    def description(self) -> str:
        return f"Write file to {self.path_template}"

    def check_before_run(self, inputs: dict) -> bool:
        if re.match(r"\{outputs.*?\}", self.path_template):
            return True
        path = self.path_template.format(inputs=inputs)
        path_obj = Path(path)
        return not path_obj.exists()

    def check_after_run(self, inputs: dict, outputs: dict) -> bool:
        path = self.path_template.format(
            inputs=inputs, outputs=outputs)
        path_obj = Path(path)
        return path_obj.exists()



import typing as T
import json


class DescriptionJSONEncoder(json.JSONEncoder):
    def default(self, o: T.Any) -> T.Any:
        from ..desc import Description, Value, SideEffect, _NotDef
        if isinstance(o, Description):
            return {
                "inputs": [self.default(v) for v in o.inputs],
                "outputs": [self.default(v) for v in o.outputs],
                "side_effects": [self.default(v) for v in o.side_effects],
            }
        elif isinstance(o, Value):
            return {
                "type": o.type.__name__ if o.type is not None else o.type,
                "range": o.range,
                "default": o.default,
            }
        elif isinstance(o, _NotDef):
            return "not_defined"
        elif isinstance(o, SideEffect):
            return {
                "type": o.__class__.__name__,
                "description": o.description,
            }
        else:
            return super().default(o)



import typing as T
from copy import copy

from .utils.json import DescriptionJSONEncoder
from .utils.misc import CreateByGetItem


class _NotDef:
    def __str__(self):
        return "NotDef"


NotDef = _NotDef()


T1 = T.TypeVar("T1")


TypeChecker = T.Callable[[T.Any, type], bool]
RangeChecker = T.Callable[[T.Any, T.Any], bool]


class Value(metaclass=CreateByGetItem):
    type_to_range_checker: T.Dict[type, TypeChecker] = {}
    type_to_type_checker: T.Dict[type, RangeChecker] = {}

    def __init__(
            self,
            type_: T.Optional[T.Type[T1]] = None,
            range_: T.Optional[T.Any] = None,
            default: T.Union[_NotDef, T1] = NotDef,
            name: str = "?",
            **kwargs,
            ):
        self.name = name
        self.type = type_
        self.range = range_
        self.default = default
        self.kwargs = kwargs

    @property
    def type_checker(self):
        if self.type is None:
            return None
        else:
            return self.type_to_type_checker.get(self.type.__name__, None)

    @property
    def range_checker(self):
        if self.type is None:
            return None
        else:
            return self.type_to_range_checker.get(self.type.__name__, None)

    @classmethod
    def register_range_check(cls, type, checker):
        cls.type_to_range_checker[type.__name__] = checker

    @classmethod
    def register_type_check(cls, type, checker=None):
        checker = checker or (lambda v, t: isinstance(v, t))
        cls.type_to_type_checker[type.__name__] = checker

    def check_type(self, val):
        if self.type_checker is not None:
            if not self.type_checker(val, self.type):
                raise TypeError(
                    f"Value {val} is not in valid type({self.type})")

    def check_range(self, val):
        if (self.range_checker is not None):
            if (not self.range_checker(val, self.range)):
                raise ValueError(
                    f"Value {val} is not in a valid range({self.range}).")

    def __repr__(self):
        return (
            f"<Value type={self.type} range={self.range} "
            f"default={self.default}>"
        )


# register basic types
def _check_number_in_range(v, range):
    return (range is None) or (range[0] <= v <= range[1])


Value.register_type_check(str)
Value.register_range_check(int, _check_number_in_range)
Value.register_type_check(int)
Value.register_range_check(float, _check_number_in_range)
Value.register_type_check(float)
Value.register_type_check(bool)


class SideEffect():
    def __init__(self, description: str):
        self._description = description

    @property
    def description(self) -> str:
        return self._description

    def check_before_run(self, inputs: dict) -> bool:
        return True

    def check_after_run(self, inputs: dict, outputs: dict) -> bool:
        return True

    def __repr__(self):
        return f"<{self.__class__.__name__} description={self.description}>"


class Description():
    def __init__(
            self,
            inputs: T.Optional[T.List[Value]] = None,
            outputs: T.Optional[T.List[Value]] = None,
            side_effects: T.Optional[T.List[SideEffect]] = None,
            ):
        self.inputs = [] if inputs is None else inputs
        self.outputs = [] if outputs is None else outputs
        self.side_effects = [] if side_effects is None else side_effects

    def parse_pass_in(self, args: tuple, kwargs: dict) -> T.Dict[str, T.Any]:
        """Get the pass in value of the func
        arguments according to the inputs description."""

        args_ = list(args)
        kwargs = copy(kwargs)
        res = {}
        for val in self.inputs:
            has_default = val.default is not NotDef
            name: str = val.name
            if len(args_) > 0:
                res[name] = args_.pop(0)
            elif (len(kwargs) > 0) and (name in kwargs):
                res[name] = kwargs.pop(name)
            else:
                if has_default:
                    res[name] = val.default
                else:
                    raise TypeError(
                        f"{name} is not provided and has no default value."
                    )
        return res

    def to_json(self) -> str:
        json_str = DescriptionJSONEncoder().encode(self)
        return json_str

    def __repr__(self) -> str:
        return (
            "<Description\n"
            f"\tinputs={self.inputs}\n"
            f"\toutputs={self.outputs}\n"
            f"\tside_effects={self.side_effects}\n"
            ">"
        )



#!/usr/bin/env python

"""The setup script."""

import re
from setuptools import setup, find_packages


def get_long_description():
    return "See https://github.com/Nanguage/funcdesc"


def get_version():
    with open("funcdesc/__init__.py") as f:
        for line in f.readlines():
            m = re.match("__version__ = '([^']+)'", line)
            if m:
                return m.group(1)
        raise IOError("Version information can not found.")


def get_install_requirements():
    requirements = [
    ]
    return requirements


requires_dev = [
    "pip", "setuptools", "wheel", "twine", "ipdb",
    'pytest', 'pytest-cov', 'flake8', 'mypy'
]


setup(
    author="Weize Xu",
    author_email='vet.xwz@gmail.com',
    python_requires='>=3.8',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    description=(
        "Establish a general function description protocol, "
        "which can realize a comprehensive description of the input, "
        "output and side effects of an target function through "
        "an Python object. Provide a unified abstraction for parameter "
        "checking, interface generation and other functions "
        "in applications such as oneFace."
    ),
    install_requires=get_install_requirements(),
    license="MIT license",
    long_description=get_long_description(),
    include_package_data=True,
    keywords='funcdesc',
    name='funcdesc',
    packages=find_packages(include=['funcdesc', 'funcdesc.*']),
    url='https://github.com/Nanguage/funcdesc',
    version=get_version(),
    zip_safe=False,
    extras_require={
        'dev': requires_dev
    }
)
